class GibError(Exception):
    pass


class MachineHeadError(GibError):
    """Raise Error when Head out of bounds"""


class InvalidInstructionError(GibError):
    """Raise Error when instruction is invalid"""



"""
Gibberish Language (gib-lang) interpreter
===============================================
3 Instructions, Reversible Turing Tarpit, Esoteric Programming Language
based on https://esolangs.org/wiki/Talk:Picofuck#Candidate_languages
"""

__version__ = "0.2.1"
__author__ = "Alifa Izzan Akhsani"
__license__ = "MIT"

from . import machine, tape, runner

Machine = machine.Machine
Tape = tape.Tape
run = runner.run


__all__ = ["Machine", "Tape", "run"]



import enum

class Instruction(enum.Enum):
    """
    Gibberish Language Syntax
    -------------------------
    FAFA : if the current command counter is 0, toggle current tape head bit then move the tape head right, toggle the command counter
    FUFU : 1. if the current command counter is 0 and current tape head bit is 0, move the machine head right until found matching FUFU
           2. if the current comand counter is 1, move the tape head left, toggle it, then move the tape head left,
              if the current tape head bit is 0, move the machine head left until found matching FUFU
           3. if the machine head is not moved in step 1 or 2, toggle the command counter

    based on https://esolangs.org/wiki/Talk:Picofuck#Candidate_languages
    """

    TOGGLE_HEAD_RIGHT = "FAFAFAFA"  # *
    HEAD_LEFT_LOOP_START = "FUFUFAFA"  # {
    LOOP_END = "FAFAFUFU"  # }
    NO_OP = "FUFUFUFU"



import pytest

from gib_lang import Machine, run
from gib_lang.machine import Instruction, MachineHeadError, InvalidInstructionError


# """
# FAFAFAFA FUFUFAFA FAFAFUFU                                   #toggle
# FAFAFAFA FUFUFAFA FAFAFUFU FAFAFAFA                          #move right
# FUFUFAFA FAFAFUFU                                            #move left
# FAFAFAFA FUFUFAFA FAFAFUFU FAFAFAFA FUFUFAFA                 #loop start
# FAFAFUFU                                                     #loop end

# """
@pytest.fixture
def forward() -> Machine:
    instruction = "FAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFAFAFAFAFAFUFUFAFAFAFAFUFU"
    return Machine(instruction)

def backward() -> Machine:
    instruction = "FAFAFAFAFUFUFAFAFAFAFUFUFUFUFAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFAFUFU"
    return Machine(instruction)

def forward_check(forward: Machine) -> None:
    machine, tape = run(forward, "100")
    
    assert machine.steps == 100
    assert tape == "010"
    
    machine.restart()
    
    machine, tape = run(machine, tape)
    
    assert machine.steps == 100
    assert tape == "001"
    
def backward_check(backward: Machine) -> None:
    machine, tape = run(backward, "100")
    
    assert machine.steps == 100
    assert tape == "010"
    
    machine.restart()
    
    machine, tape = run(machine, tape)
    
    assert machine.steps == 100
    assert tape == "100"

def test_drive() -> None:
    source = "FAFAFAFA"*8
    machine = Machine(source)

    assert machine == source

def test_hash() -> None:
    source = "FAFAFAFA"*8
    machine = Machine(source)
    
    assert hash(machine) == hash(source)

def test_eq() -> None:
    source = "FAFAFAFA"*8
    machine = Machine(source)
    machine2 = Machine(source)
    assert machine == machine2

def test_copy() -> None:
    source = "FAFAFAFA"*8
    machine = Machine(source)
    machine2 = machine.copy()
    assert machine == machine2

def test_invalid_instruction() -> None:
    source = "FAFAFAFAFAFAFUFU"
    with pytest.raises(InvalidInstructionError):
         Machine(source)
    

def test_head() -> None:
    source = "FAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFAFUFUFAFAFAFA"*8
    machine = Machine(source)

    assert len(machine) == 56
    assert machine.head == 0

    for i in range(55):
        assert machine.head == i
        machine.move_right()
    
    assert machine.steps == 55
    assert machine.head == 55
    with pytest.raises(MachineHeadError):
        machine.move_right()
    
    assert machine.head == 55
    assert machine.steps == 56

    for i in range(55, 0, -1):
        assert machine.head == i
        machine.move_left()
    assert machine.head == 0
    assert machine.steps == 111
    with pytest.raises(MachineHeadError):
        machine.move_left()
    assert machine.head == 0
    assert machine.steps == 112

def test_restart() -> None:
    source = "FAFAFAFAFAFAFAFA"*8
    machine = Machine(source)

    machine.move_right(8)
    
    assert machine.head == 8
    assert machine.steps == 8

    machine.restart()

    assert machine.head == 0
    assert machine.steps == 0

def test_getitem() -> None:
    source = "FUFUFAFAFAFAFUFU"*8
    machine = Machine(source)

    assert machine[0] == Instruction.HEAD_LEFT_LOOP_START
    assert machine[1] == Instruction.LOOP_END
    assert machine[2] == Instruction.HEAD_LEFT_LOOP_START

def test_instruction_prop() -> None:
    source = "FUFUFAFAFAFAFUFU"*8
    machine = Machine(source)

    assert machine.instruction == Instruction.HEAD_LEFT_LOOP_START
    machine.move_right()
    assert machine.instruction == Instruction.LOOP_END
    machine.move_right()
    assert machine.instruction == Instruction.HEAD_LEFT_LOOP_START

#TODO: loop tests


import argparse
import logging
import os
import sys

from . import Machine, Tape, run, __version__

pr_ascii = False

def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    parser.add_argument(
        "--log-level",
        help="The log level, default is INFO",
        default="INFO",
    )

    subparsers = parser.add_subparsers(
        required=True,
        dest="subcommand",
    )

    run_parser = subparsers.add_parser("run", help="Run a code")

    run_parser.add_argument("source", help="The code to run")
    run_parser.add_argument(
        "-t",
        "--tape",
        help="The initial tape to use. Can be an integer to use as the tape size, or a string of 1s and 0s",
        default="8",
    )
    run_parser.add_argument(
        "--max-steps",
        type=int,
        help="The maximum number of steps",
        default=10000,
    )

    run_parser.add_argument(
        "-c",
        "--ascii",
        action=argparse.BooleanOptionalAction
    )

    args = parser.parse_args()

    logger = logging.getLogger("gib")

    logger.setLevel(args.log_level.upper())
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(levelname)05s %(name)s: %(message)s"))
    logger.addHandler(handler)

    if args.subcommand == "run":
        run_main(args, logger)
    else:
        raise ValueError(f"Unknown command: {args.subcommand}")


def pipe_print(*args: object, **kwargs: object) -> None:
    try:
        print(*args, **kwargs)  # type: ignore
        sys.stdout.flush()
    except BrokenPipeError:
        # Gracefully handle broken pipe when e.g. piping to head
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        sys.exit(1)

def _preprocess(source: str) -> str:
    source = source.replace("\\n", "\n")
    raws = source.split("\n")
    raws = [raw.split("#", 1)[0] for raw in raws]
    source = "".join(raws)
    source = source.replace(" ", "")
    return source

def process_file(path: str) -> str:
    try:
        with open(path, 'r') as f:
            content = f.read()
            return _preprocess(content)
    except FileNotFoundError:
        print(f"Error: The file '{path}' was not found.")
    except Exception as e:
        print(f"An error occurred: {e}")
    return ""

def run_main(args: argparse.Namespace, logger: logging.Logger) -> None:
    # Check if tape is a string containing only 1s and 0s
    if all(x in "01" for x in args.tape):
        logger.debug("Using --tape as a string")
        tape = args.tape
    else:
        logger.debug("Using --tape as an integer")
        tape = int(args.tape)

    def callback(machine: Machine, tape: Tape) -> bool:
        logger.debug(
            f" {machine.steps} {machine.instruction.value} {machine.head:02d} | {tape.head:02d} {tape}"
        )
        return False
    source = ""
    if args.source.find('/') != -1:
        source = process_file(args.source)
    else:
        source = args.source

    _machine, tape = run(
        _preprocess(source),
        tape,
        max_steps=args.max_steps,
        callback=callback,
    )

    if args.ascii:
        pipe_print(tape.to_ascii())
    else:
        pipe_print(tape)


if __name__ == "__main__":
    main()



import logging
from typing import Callable, Optional

from .instruction import Instruction
from .tape import Tape, _TapeInitType
from .machine import Machine, _MachineInitType, MachineHeadError

logger = logging.getLogger("gib")


def run(
    machine: _MachineInitType,
    tape: _TapeInitType,
    max_steps: int = 100,
    callback: Optional[Callable[[Machine, Tape], bool]] = None,
) -> tuple[Machine, Tape]:
    """Run a Gibberish Language program on a tape"""
    machine = Machine(machine)
    tape = Tape(tape)
    try:
        while machine.steps < max_steps:
            if callback and callback(machine, tape):
                break
            instruction = machine.instruction
            current_bit = tape.get
            if instruction == Instruction.TOGGLE_HEAD_RIGHT:
                tape.toggle()
                tape.move_right()
                machine.move_right()
            elif instruction == Instruction.HEAD_LEFT_LOOP_START:
                tape.move_left()
                logger.debug(f"Loop start at head {machine.head}, current bit: {current_bit}")
                machine.loop_start(current_bit)
            elif instruction == Instruction.LOOP_END:
                logger.debug(f"Loop end at head {machine.head}, current bit: {current_bit}")
                machine.loop_end(current_bit)
            elif instruction == Instruction.NO_OP:
                machine.move_right()
            else:
                raise ValueError(f"Unknown instruction: {instruction}")

    except MachineHeadError:
        pass
    return machine, tape



import pytest
from gib_lang import Tape

@pytest.fixture
def tape()-> Tape:
    return Tape(8)

def test_tape(tape: Tape) -> None:
    assert len(tape) == 8
    assert tape.head == 0
    assert tape.tape == [False]*8

def test_getitem(tape: Tape) -> None:
    assert tape[0] == False
    assert tape[1:4] == [False, False, False]

def test_copy(tape: Tape) -> None:
    tape2 = tape.copy()
    assert tape == tape2
    assert tape is not tape2

    tape3 = Tape(tape)
    assert tape == tape3
    assert tape is not tape3

def test_hashable(tape: Tape) -> None:
    tape2= tape.copy()
    assert hash(tape) == hash(tape2)
    assert hash(tape) == hash("00000000")

def test_init() -> None:
    tape1 = Tape(5)
    assert tape1.tape == [False, False, False, False, False]

    tape2 = Tape("10101")
    assert tape2.tape == [True, False, True, False, True]

    tape3 = Tape([1, 0, 1, 0])
    assert tape3.tape == [True, False, True, False]

def test_toggle(tape: Tape) -> None:
    assert tape.get == False
    tape._tape[tape.head] = not tape.get
    assert tape.get == True
    tape.toggle()
    assert tape.get == False

def test_reset(tape: Tape) -> None:
    tape.move_right(3)
    tape.toggle()
    assert tape.head == 3
    assert tape.get == True
    tape.reset()
    assert tape.head == 0
    assert tape.tape == [False]*8

def test_head(tape: Tape) -> None:
    assert tape.head == 0

    for i in range(7):
        assert tape.head == i
        tape.move_right()
    assert tape.head == 7
    tape.move_right()
    assert tape.head == 0
    tape.move_left()
    assert tape.head == 7
    for i in range(7, 0, -1):
        assert tape.head == i
        tape.move_left()


from gib_lang.runner import Machine, Tape, run

def test_run_toggle() -> None:
    source = "FAFAFAFA" * 8
    tape_size = 8

    machine, tape = run(
        source,
        tape_size,
    )

    assert machine == source
    assert tape == "11111111"

    source = "FAFAFAFAFUFUFAFAFUFUFAFAFAFAFUFUFUFUFAFAFAFAFUFUFAFAFUFU" * 8 
    machine, tape = run(
        source,
        tape,
        1000,
        callback=callback,
    )

    assert machine == source
    assert tape == "01010101"


def test_run_callback() -> None:
    source = "FAFAFAFA" * 8
    tape_size = 8

    def callback(machine: Machine, tape: Tape) -> bool:
        return machine.steps == 5

    machine, tape = run(
        source,
        tape_size,
        callback=callback,
    )

    assert machine.steps == 5
    assert tape == "11111000"


def test_run_max_steps() -> None:
    source = "FAFAFAFA" * 8
    tape_size = 8

    machine, tape = run(
        source,
        tape_size,
        max_steps=4,
    )

    assert machine.steps == 4
    assert tape == "11110000"


def test_run_loop_behavior() -> None:
    source = "FUFUFAFAFAFAFUFU"
    machine, tape = run(source, 8, max_steps=10)
    
    assert machine.steps == 1
    assert tape == "00000000"

    source = "FAFAFAFAFUFUFAFAFUFUFAFAFAFAFUFUFUFUFAFAFAFAFUFUFAFAFUFU"*8
    machine, tape = run(source, 8, max_steps=10)
    assert machine.steps == 10


from collections.abc import Sequence
from typing import Union, Optional, overload
import warnings

_TAPE_LENGHT = 8
_TapeInitType = Union[int, str, Sequence, "Tape"]


class Tape(Sequence[bool]):
    """Circular Tape"""

    _tape: list[bool]
    _head: int

    def __init__(
        self,
        tape: _TapeInitType = _TAPE_LENGHT,
        head: Optional[int] = None,
    ) -> None:
        if isinstance(tape, int):
            self._tape = [False] * tape
        elif isinstance(tape, str):
            self._tape = [bool(int(x)) for x in tape]
        elif isinstance(tape, Tape):
            self._tape = tape._tape.copy()
            if head is not None:
                warnings.warn(
                    "Head argument is ignored when initializing Tape with "
                    "another Tape. Set it to None to disable this warning.",
                    stacklevel=2,
                )
            head = tape._head
        elif isinstance(tape, Sequence):
            # Initialize the tape with the given sequence
            self._tape = [bool(x) for x in tape]
        else:
            raise TypeError("Tape must be initialized with an int or a sequence.")

        self._head = 0 if head is None else head

    @property
    def tape(self) -> Sequence[bool]:
        # Return a copy of the tape
        return list(self._tape)

    @property
    def get(self) -> bool:
        return self._tape[self._head]

    @property
    def head(self) -> int:
        return self._head

    def __len__(self) -> int:
        return len(self._tape)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Tape):
            return self._tape == other._tape
        elif isinstance(other, str):
            return str(self) == other
        elif isinstance(other, Sequence):
            return self._tape == [bool(x) for x in other]
        else:
            return False

    @overload
    def __getitem__(self, index: int) -> bool: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[bool]: ...

    def __getitem__(
        self, index_or_slice: Union[int, slice]
    ) -> Union[bool, Sequence[bool]]:
        if isinstance(index_or_slice, (int, slice)):
            return self._tape[index_or_slice]
        else:
            raise TypeError("Index must be an int or a slice.")

    def _single_char_rp(self) -> str:
        return "".join("1" if bit else "0" for bit in self._tape)

    def __repr__(self) -> str:
        return f"Tape({self._single_char_rp()!r})"

    def __str__(self) -> str:
        return self._single_char_rp()

    def __hash__(self) -> int:
        return hash(str(self))

    def copy(self) -> "Tape":
        """Return a copy of the tape."""
        return Tape(self)

    def reset(self) -> None:
        """Reset the tape to all 0s and move the head to the start."""
        self._tape = [False] * len(self)
        self._head = 0

    def toggle(self) -> None:
        """Toggle the current cell."""
        self._tape[self._head] = not self._tape[self._head]

    def move_right(self, n: int = 1) -> None:
        """Move to the right."""
        self._head = (self._head + n) % len(self)

    def move_left(self, n: int = 1) -> None:
        """Move to the left."""
        self._head = (self._head - n) % len(self)

    def to_ascii(self) -> str:
        """Convert the tape to an ASCII string."""
        if len(self) % 8 != 0:
            raise ValueError("Tape length must be a multiple of 8 to convert to ASCII.")
        chars = []
        for i in range(0, len(self), 8):
            byte = self[i : i + 8]
            byte_str = "".join("1" if bit else "0" for bit in byte)
            chars.append(chr(int(byte_str, 2)))
        return "".join(chars)



from collections.abc import Sequence

from typing import Union, Optional, overload
import warnings

from .instruction import Instruction
from .exceptions import MachineHeadError, InvalidInstructionError

_MachineInitType = Union[str, Sequence[Instruction], "Machine"]


class Machine(Sequence[Instruction]):
    """Gibberish Language Machine based on RBF"""

    _machine: list[Instruction]
    _head: int
    _steps: int

    def __init__(
        self,
        machine: _MachineInitType,
        head: Optional[int] = None,
    ) -> None:
        if isinstance(machine, Machine):
            self._machine = list(machine.machine)
            self._head = machine.head
            self._steps = machine.steps
            if head is not None:
                warnings.warn(
                    "head parameter is ignored when initializing from another Machine instance",
                    stacklevel=2,
                )
        elif isinstance(machine, (str, Sequence)):
            validated_machine = validate_machine(machine)
            self._machine = validated_machine
            head = 0 if head is None else head
            self._head = head
            self._steps = 0
        else:
            raise TypeError(
                "Machine must be initialized with a string, a sequence of Instructions, or another Machine instance"
            )

    @property
    def machine(self) -> Sequence[Instruction]:
        """returns a copy of the machine's instructions"""
        return self._machine.copy()

    @property
    def head(self) -> int:
        """returns the current head position"""
        return self._head

    @property
    def steps(self) -> int:
        """returns the number of steps taken"""
        return self._steps

    @property
    def instruction(self) -> Instruction:
        if len(self) == 0:
            raise MachineHeadError("Machine is empty")
        return self._machine[self._head]

    def __len__(self) -> int:
        return len(self._machine)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Machine):
            return self._machine == other.machine
        elif isinstance(other, str):
            return str(self) == other
        else:
            return False

    @overload
    def __getitem__(self, index: int) -> Instruction: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[Instruction]: ...

    def __getitem__(
        self,
        index_or_slice: Union[int, slice],
    ) -> Union[Instruction, Sequence[Instruction]]:
        if isinstance(index_or_slice, (int, slice)):
            return self._machine[index_or_slice]
        else:
            raise TypeError("Index must be an integer or a slice")

    def _single_char_rp(self) -> str:
        return "".join(instruction.value for instruction in self._machine)

    def __repr__(self) -> str:
        return f"Machine({self._single_char_rp()!r})"

    def __str__(self) -> str:
        return self._single_char_rp()

    def __hash__(self) -> int:
        return hash(str(self))

    def copy(self) -> "Machine":
        return Machine(self)

    def restart(self) -> None:
        self._head = 0
        self._steps = 0

    def _move_right(self) -> None:
        if self._head < len(self) - 1:
            self._head += 1
        else:
            raise MachineHeadError("Machine head overflow")

    def move_right(self, n: int = 1) -> None:
        for _ in range(n):
            self._steps += 1
            self._move_right()

    def _move_left(self) -> None:
        if self._head > 0:
            self._head -= 1
        else:
            raise MachineHeadError("Machine head underflow")

    def move_left(self, n: int = 1) -> None:
        for _ in range(n):
            self._steps += 1
            self._move_left()

    def loop_start(self, current_bit: bool) -> None:
        if self.instruction != Instruction.HEAD_LEFT_LOOP_START:
            raise ValueError("Not at a loop start.")

        self._steps += 1
        
        if not current_bit:
            try:
                bracket_depth = 1
                while bracket_depth > 0:
                    self._move_right()
                    if self.instruction == Instruction.HEAD_LEFT_LOOP_START:
                        bracket_depth += 1
                    elif self.instruction == Instruction.LOOP_END:
                        bracket_depth -= 1

            except MachineHeadError:
                raise InvalidInstructionError("Unmatched loop start.") from None

        try:
            self._move_right()
        except MachineHeadError:
            raise

    def loop_end(self, current_bit: bool) -> None:
        if self.instruction != Instruction.LOOP_END:
            raise ValueError("Not at a loop end.")

        self._steps += 1

        if not current_bit:
            try:
                bracket_depth = 1
                while bracket_depth > 0:
                    self._move_left()
                    if self._machine[self._head] == Instruction.HEAD_LEFT_LOOP_START:
                        bracket_depth -= 1
                    elif self._machine[self._head] == Instruction.LOOP_END:
                        bracket_depth += 1

            except MachineHeadError:
                raise InvalidInstructionError("Unmatched loop end.") from None
        self._move_right()

def _chunk_string(s, n):
    return (s[i:i+n] for i in range(0, len(s), n))

def validate_machine(machine: Union[str, Sequence[Instruction]]) -> list[Instruction]:
    valid_instructions: Sequence[Instruction]
    if isinstance(machine, str):
        try:

            valid_instructions = [Instruction(instruction) for instruction in _chunk_string(machine, 8)]
        except ValueError as e:
            raise InvalidInstructionError(str(e)) from None
    else:
        valid_instructions = list(machine)

    depth = 0
    for instruction in valid_instructions:
        if instruction == Instruction.HEAD_LEFT_LOOP_START:
            depth += 1
        elif instruction == Instruction.LOOP_END:
            depth -= 1

        if depth < 0:
            raise InvalidInstructionError(
                "Invalid instruction sequence: unexpected FAFAFUFU"
            )
    if depth != 0:
        raise InvalidInstructionError(
            "Invalid instruction sequence: unmatched FUFUFAFA "
        )

    return valid_instructions
